import os
import shutil
import subprocess
from contextlib import suppress

import pytest
from astropy import units as u
from astropy.io.fits import Header

from panoptes.utils.images import fits as fits_utils


@pytest.fixture
def solved_fits_file(data_dir):
    return os.path.join(data_dir, 'solved.fits.fz')


def test_wcsinfo(solved_fits_file):
    wcsinfo = fits_utils.get_wcsinfo(solved_fits_file)

    assert 'wcs_file' in wcsinfo
    assert wcsinfo['ra_center'].value == 303.206422334


def test_fpack(solved_fits_file):
    new_file = solved_fits_file.replace('solved', 'solved_copy')
    copy_file = shutil.copyfile(solved_fits_file, new_file)
    info = os.stat(copy_file)
    assert info.st_size > 0.

    uncompressed = fits_utils.funpack(copy_file)
    assert os.stat(uncompressed).st_size > info.st_size

    compressed = fits_utils.fpack(uncompressed)
    assert os.stat(compressed).st_size == info.st_size

    os.remove(copy_file)


def test_no_overwrite_fpack(solved_fits_file):
    new_file = solved_fits_file.replace('solved', 'solved_copy')
    copy_file = shutil.copyfile(solved_fits_file, new_file)

    # Unpack the file. This removes the packed version.
    uncompressed = fits_utils.funpack(copy_file)

    # Copy file again so now the packed version exists alongside unpacked.
    copy_file = shutil.copyfile(solved_fits_file, new_file)

    # Deny overwriting gives error.
    with pytest.raises(FileExistsError):
        _ = fits_utils.fpack(uncompressed, overwrite=False)

    # Default is overwrite=True.
    compressed = fits_utils.fpack(uncompressed)

    # Cleanup test.
    for file in [copy_file, uncompressed, compressed]:
        with suppress(FileNotFoundError):
            os.remove(file)


def test_getheader(solved_fits_file):
    header = fits_utils.getheader(solved_fits_file)
    assert isinstance(header, Header)
    assert header['IMAGEID'] == 'PAN001_XXXXXX_20160909T081152'


def test_getval(solved_fits_file):
    img_id = fits_utils.getval(solved_fits_file, 'IMAGEID')
    assert img_id == 'PAN001_XXXXXX_20160909T081152'


def test_solve_field_unsolved(unsolved_fits_file):
    with pytest.raises(KeyError):
        fits_utils.getval(unsolved_fits_file, 'WCSAXES')

    assert 'crpix0' not in fits_utils.get_wcsinfo(unsolved_fits_file)

    proc = fits_utils.solve_field(unsolved_fits_file)
    assert isinstance(proc, subprocess.Popen)
    proc.wait()
    try:
        outs, errs = proc.communicate(timeout=15)
    except subprocess.TimeoutExpired:
        proc.kill()
        outs, errs = proc.communicate()

    assert proc.returncode == 0, f'{outs}\n{errs}'

    wcs_info = fits_utils.get_wcsinfo(unsolved_fits_file.replace('.fits', '.new'))
    assert 'crpix0' in wcs_info
    assert wcs_info['crpix0'] == 350.5 * u.pixel

    for ext in ['.wcs', '.new']:
        with suppress(FileNotFoundError):
            os.remove(unsolved_fits_file.replace('.fits', ext))


def test_get_solve_field_solved(solved_fits_file):
    orig_wcs = fits_utils.get_wcsinfo(solved_fits_file)
    assert 'crpix0' in orig_wcs

    solve_info = fits_utils.get_solve_field(solved_fits_file, skip_solved=False)
    assert isinstance(solve_info, dict)
    # 1-based numbering from WCS.
    assert 'CRPIX1' in solve_info


def test_solve_options(solved_fits_file):
    proc = fits_utils.solve_field(
        solved_fits_file, solve_opts=['--guess-scale'])
    assert isinstance(proc, subprocess.Popen)
    proc.wait()
    assert proc.returncode == 0


def test_solve_bad_field(solved_fits_file):
    proc = fits_utils.solve_field('Foo.fits')
    outs, errs = proc.communicate()
    print('outs', outs)
    print('errs', errs)
    assert 'ERROR' in errs





def focus_metric(data, merit_function='vollath_F4', **kwargs):
    """Compute the focus metric.

    Computes a focus metric on the given data using a supplied merit function.
    The merit function can be passed either as the name of the function (must be
    defined in this module) or as a callable object. Additional keyword arguments
    for the merit function can be passed as keyword arguments to this function.

    Args:
        data (numpy array) -- 2D array to calculate the focus metric for.
        merit_function (str/callable) -- Name of merit function (if in
            panoptes.utils.images) or a callable object.

    Returns:
        scalar: result of calling merit function on data
    """
    if isinstance(merit_function, str):
        try:
            merit_function = globals()[merit_function]
        except KeyError:
            raise KeyError(
                "Focus merit function '{}' not found in panoptes.utils.images!".format(merit_function))

    return merit_function(data, **kwargs)


def vollath_F4(data, axis=None):
    """Compute F4 focus metric

    Computes the F_4 focus metric as defined by Vollath (1998) for the given 2D
    numpy array. The metric can be computed in the y axis, x axis, or the mean of
    the two (default).

    Arguments:
        data (numpy array) -- 2D array to calculate F4 on.
        axis (str, optional, default None) -- Which axis to calculate F4 in. Can
            be 'Y'/'y', 'X'/'x' or None, which will calculate the F4 value for
            both axes and return the mean.

    Returns:
        float64: Calculated F4 value for y, x axis or both
    """
    def _vollath_F4_y():
        A1 = (data[1:] * data[:-1]).mean()
        A2 = (data[2:] * data[:-2]).mean()
        return A1 - A2

    def _vollath_F4_x():
        A1 = (data[:, 1:] * data[:, :-1]).mean()
        A2 = (data[:, 2:] * data[:, :-2]).mean()
        return A1 - A2

    if str(axis).lower() == 'y':
        return _vollath_F4_y()
    elif str(axis).lower() == 'x':
        return _vollath_F4_x()
    elif not axis:
        return (_vollath_F4_y() + _vollath_F4_x()) / 2
    else:
        raise ValueError(f"axis must be one of 'Y', 'y', 'X', 'x' or None, got {axis}!")



"""Provides SerialData, a PySerial wrapper."""

import operator
import serial
from serial.tools.list_ports import comports as get_comports
import time
from contextlib import suppress

from . import error
from .logger import logger
from .serializers import from_json


# Note: get_serial_port_info is replaced by tests to override the normal
# behavior, so don't change the name without fixing the tests.
def get_serial_port_info():
    """Returns the serial ports defined on the system.

    Returns: a list of PySerial's ListPortInfo objects. See:
        https://github.com/pyserial/pyserial/blob/master/serial/tools/list_ports_common.py
    """
    return sorted(get_comports(), key=operator.attrgetter('device'))


class SerialData(object):
    """SerialData wraps a PySerial instance for reading from and writing to a serial device.

    Because POCS is intended to be very long running, and hardware may be turned off when unused
    or to force a reset, this wrapper may or may not have an open connection to the underlying
    serial device. Note that for most devices, is_connected will return true if the device is
    turned off/unplugged after a connection is opened; the code will only discover there is a
    problem when we attempt to interact with the device.

    .. doctest::

        >>> import serial

        # Register our serial simulators
        >>> serial.protocol_handler_packages.append('panoptes.utils.tests.serial_handlers')
        >>> from panoptes.utils.tests.serial_handlers import protocol_buffers as pb

        # Import our serial utils
        >>> from panoptes.utils.rs232 import SerialData

        # Connect to our fake buffered device
        >>> device_listener = SerialData(port='buffers://')

        # Note: A manual reset is currently required because implementation is not complete.
        # See https://github.com/panoptes/POCS/issues/758 for details.
        >>> pb.ResetBuffers()
        >>> device_listener.is_connected
        True

        >>> device_listener.port
        'buffers://'

        # Device sends event
        >>> pb.SetRBufferValue(b'emit event')

        # Listen for event
        >>> device_listener.read()
        'emit event'

        >>> device_listener.write('ack event')
        9
        >>> pb.GetWBufferValue()
        b'ack event'

        # Remove custom handlers
        >>> serial.protocol_handler_packages.remove('panoptes.utils.tests.serial_handlers')
    """

    def __init__(self,
                 port=None,
                 baudrate=115200,
                 name=None,
                 timeout=2.0,
                 open_delay=0.0,
                 retry_limit=5,
                 retry_delay=0.5,
                 **kwargs
                 ):
        """Create a SerialData instance and attempt to open a connection.

        The device need not exist at the time this is called, in which case is_connected will
        be false.

        Args:
            port: The port (e.g. /dev/tty123 or socket://host:port) to which to
                open a connection.
            baudrate: For true serial lines (e.g. RS-232), sets the baud rate of
                the device.
            name: Name of this object. Defaults to the name of the port.
            timeout (float, optional): Timeout in seconds for both read and write.
                Defaults to 2.0.
            open_delay: Seconds to wait after opening the port.
            retry_limit: Number of times to try readline() calls in read().
            retry_delay: Delay between readline() calls in read().

        Raises:
            ValueError: If the serial parameters are invalid (e.g. a negative baudrate).

        """
        self.logger = logger

        if not port:
            raise ValueError('Must specify port for SerialData')

        self.name = name or port
        self.retry_limit = retry_limit
        self.retry_delay = retry_delay

        self.ser = serial.serial_for_url(port, do_not_open=True)

        # Configure the PySerial class.
        self.ser.baudrate = baudrate
        self.ser.bytesize = serial.EIGHTBITS
        self.ser.parity = serial.PARITY_NONE
        self.ser.stopbits = serial.STOPBITS_ONE
        self.ser.timeout = timeout
        self.ser.write_timeout = timeout
        self.ser.xonxoff = False
        self.ser.rtscts = False
        self.ser.dsrdtr = False

        self.logger.debug('SerialData for {} created', self.name)

        # Properties have been set to reasonable values, ready to open the port.
        try:
            self.ser.open()
        except serial.serialutil.SerialException as err:
            self.logger.debug('Unable to open {}. Error: {}', self.name, err)
            return

        open_delay = max(0.0, float(open_delay))
        if open_delay > 0.0:
            self.logger.debug('Opened {}, sleeping for {} seconds', self.name, open_delay)
            time.sleep(open_delay)
        else:
            self.logger.debug('Opened {}', self.name)

    @property
    def port(self):
        """Name of the port."""
        return self.ser.port

    @property
    def is_connected(self):
        """True if serial port is open, False otherwise."""
        return self.ser.is_open

    def connect(self):
        """If disconnected, then connect to the serial port.

        Raises:
            error.BadSerialConnection if unable to open the connection.
        """
        if self.is_connected:
            self.logger.debug('Connection already open to {}', self.name)
            return
        self.logger.debug('SerialData.connect called for {}', self.name)
        try:
            # Note: we must not call open when it is already open, else an exception is thrown of
            # the same type thrown when open fails to actually open the device.
            self.ser.open()
            if not self.is_connected:
                raise error.BadSerialConnection(
                    msg="Serial connection {} is not open".format(self.name))
        except serial.serialutil.SerialException as err:
            raise error.BadSerialConnection(msg=err)
        self.logger.debug('Serial connection established to {}', self.name)

    def disconnect(self):
        """Closes the serial connection.

        Raises:
            error.BadSerialConnection if unable to close the connection.
        """
        # Fortunately, close() doesn't throw an exception if already closed.
        self.logger.debug('SerialData.disconnect called for {}', self.name)
        try:
            self.ser.close()
        except Exception as err:
            raise error.BadSerialConnection(
                msg="SerialData.disconnect failed for {}; underlying error: {}".format(
                    self.name, err))
        if self.is_connected:
            raise error.BadSerialConnection(
                msg="SerialData.disconnect failed for {}".format(self.name))

    def write_bytes(self, data):
        """Write data of type bytes."""
        assert self.ser
        assert self.ser.isOpen()
        return self.ser.write(data)

    def write(self, value):
        """Write value (a string) after encoding as bytes."""
        return self.write_bytes(value.encode())

    def read_bytes(self, size=1):
        """Reads size bytes from the serial port.

        If a read timeout is set on self.ser, this may return less characters than requested.
        With no timeout it will block until the requested number of bytes is read.

        Args:
            size: Number of bytes to read.
        Returns:
            Bytes read from the port.
        """
        assert self.ser
        assert self.ser.isOpen()
        return self.ser.read(size=size)

    def read(self, retry_limit=None, retry_delay=None):
        """Reads next line of input using readline.

        If no response is given, delay for retry_delay and then try to read
        again. Fail after retry_limit attempts.
        """
        assert self.ser
        assert self.ser.isOpen()

        if retry_limit is None:
            retry_limit = self.retry_limit
        if retry_delay is None:
            retry_delay = self.retry_delay

        for _ in range(retry_limit):
            data = self.ser.readline()
            if data:
                return data.decode(encoding='ascii')
            time.sleep(retry_delay)
        return ''

    def get_reading(self):
        """Reads and returns a line, along with the timestamp of the read.

        Returns:
            A pair (tuple) of (timestamp, line). The timestamp is the time of completion of the
            readline operation.
        """
        # Get the timestamp after the read so that a long delay on reading doesn't make it
        # appear that the read happened much earlier than it did.
        line = self.read()
        ts = time.strftime('%Y-%m-%dT%H:%M:%S %Z', time.gmtime())
        info = (ts, line)
        return info

    def get_and_parse_reading(self, retry_limit=5):
        """Reads a line of JSON text and returns the decoded value, along with the current time.

        Args:
            retry_limit: Number of lines to read in an attempt to get one that parses as JSON.

        Returns:
            A pair (tuple) of (timestamp, decoded JSON line). The timestamp is the time of
            completion of the readline operation.
        """
        for _ in range(max(1, retry_limit)):
            (ts, line) = self.get_reading()
            if not line:
                continue

            with suppress(error.InvalidDeserialization):
                data = from_json(line)
                if data:
                    return (ts, data)
        return None

    def reset_input_buffer(self):
        """Clear buffered data from connected port/device.

        Note that Wilfred reports that the input from an Arduino can seriously lag behind
        realtime (e.g. 10 seconds), and that clear_buffer may exist for that reason (i.e. toss
        out any buffered input from a device, and then read the next full line, which likely
        requires tossing out a fragment of a line).
        """
        self.ser.reset_input_buffer()

    def __del__(self):
        """Close the serial device on delete.

        This is to avoid leaving a file or device open if there are multiple references
        to the serial.Serial object.
        """
        try:
            # If an exception is thrown when running __init__, then self.ser may not have
            # been set, in which case reading that field will generate a AttributeError.
            ser = self.ser
        except AttributeError:
            return
        if ser and ser.is_open:
            ser.close()



import socket

from .logger import logger
from . import error


class TheSkyX(object):

    """ A socket connection for communicating with TheSkyX


    """

    def __init__(self, host='localhost', port=3040, connect=True, *args, **kwargs):
        self.logger = logger

        self._host = host
        self._port = port

        self._socket = None

        self._is_connected = False
        if connect:
            self.connect()

    @property
    def is_connected(self):
        return self._is_connected

    def connect(self):
        """ Sets up serial connection """
        self.logger.debug('Making TheSkyX connection at {}:{}'.format(self._host, self._port))
        if not self.is_connected:

            try:
                self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.socket.connect((self._host, self._port))
            except ConnectionRefusedError:
                self.logger.warning('Cannot create connection to TheSkyX')
            else:
                self._is_connected = True
                self.logger.info('Connected to TheSkyX via {}:{}'.format(self._host, self._port))

    def write(self, value):
        try:
            assert isinstance(value, str)
            self.socket.sendall(value.encode())
        except AttributeError:
            raise error.BadConnection("Not connected to TheSkyX")

    def read(self, timeout=5):
        try:
            self.socket.settimeout(timeout)
            response = None
            err = None

            try:
                response = self.socket.recv(2048).decode()
                if '|' in response:
                    response, err = response.split('|')

                if 'Error:' in response:
                    response, err = response.split(':')

                if err is not None and 'No error' not in err:
                    if 'Error = 303' in err:
                        raise error.TheSkyXKeyError("Invalid TheSkyX key")

                    raise error.TheSkyXError(err)
            except socket.timeout:  # pragma: no cover
                raise error.TheSkyXTimeout()

            return response
        except AttributeError:
            raise error.BadConnection("Not connected to TheSkyX")



from contextlib import suppress
from copy import deepcopy
from collections import OrderedDict
import json
from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO

import numpy as np
from astropy.time import Time
from astropy import units as u

from . import error


class StringYAML(YAML):
    def dump(self, data, stream=None, **kwargs):
        """YAML class that can dump to a string.

        By default the YAML parser doesn't serialize directly to a string. This
        class is a small wrapper to output StreamIO as a string if no stream is
        provided.

        See https://yaml.readthedocs.io/en/latest/example.html#output-of-dump-as-a-string.

        Note:

            This class should not be used directly but instead is instantiated as
            part of the yaml convenience methods below.

        Args:
            data (`object`): An object, usually dict-like.
            stream (`None` | stream, optional): A stream object to write the YAML.
                If default `None`, return value as string.
            **kwargs: Keywords passed to the `dump` function.

        Returns:
            `str`: The serialized object string.
        """
        inefficient = False
        if stream is None:
            inefficient = True
            stream = StringIO()
        yaml = YAML()
        yaml.dump(data, stream, **kwargs)
        if inefficient:
            return stream.getvalue()


def to_json(obj, filename=None, append=True, **kwargs):
    """Convert a Python object to a JSON string.

    Will handle `datetime` objects as well as `astropy.unit.Quantity` objects.
    Astropy quantities will be converted to a dict: `{"value": val, "unit": unit}`.

    Examples:
        .. doctest::

            >>> from panoptes.utils.serializers import to_json
            >>> from astropy import units as u
            >>> config = { "name": "Mauna Loa", "elevation": 3397 * u.meter }
            >>> to_json(config)
            '{"name": "Mauna Loa", "elevation": "3397.0 m"}'

            >>> to_json({"numpy_array": np.arange(10)})
            '{"numpy_array": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]}'

            >>> from panoptes.utils import current_time
            >>> to_json({"current_time": current_time()})       # doctest: +SKIP
            '{"current_time": "2019-04-08 22:19:28.402198"}'

    Args:
        obj (`object`): The object to be converted to JSON, usually a dict.
        filename (str, optional): Path to file for saving.
        append (bool, optional): Append to `filename`, default True. Setting
            False will clobber the file.
        **kwargs: Keyword arguments passed to `json.dumps`.

    Returns:
        `str`: The JSON string representation of the object.
    """
    json_str = json.dumps(obj, default=_serialize_object, **kwargs)

    if filename is not None:
        mode = 'w'
        if append:
            mode = 'a'
        with open(filename, mode) as fn:
            fn.write(json_str + '\n')

    return json_str


def from_json(msg):
    """Convert a JSON string into a Python object.

    Astropy quanitites will be converted from a ``{"value": val, "unit": unit}``
    format. Additionally, the following units will be converted if the value ends
    with the exact string:

        * deg
        * m
        * s

    Time-like values are *not* parsed, however see example below.

    Examples:

        .. doctest::

            >>> from panoptes.utils.serializers import from_json
            >>> config_str = '{"name":"Mauna Loa","elevation":{"value":3397.0,"unit":"m"}}'
            >>> from_json(config_str)
            {'name': 'Mauna Loa', 'elevation': <Quantity 3397. m>}

            # Invalid values will be returned as is.
            >>> from_json('{"horizon":{"value":42.0,"unit":"degr"}}')
            {'horizon': {'value': 42.0, 'unit': 'degr'}}

            # The following will convert if final string:
            >>> from_json('{"horizon": "42.0 deg"}')
            {'horizon': <Quantity 42. deg>}

            >>> from_json('{"elevation": "1000 m"}')
            {'elevation': <Quantity 1000. m>}

            >>> from_json('{"readout_time": "10 s"}')
            {'readout_time': <Quantity 10. s>}

            # Be careful with short unit names in extended format!
            >>> horizon = from_json('{"horizon":{"value":42.0,"unit":"d"}}')
            >>> horizon['horizon']
            <Quantity 42. d>
            >>> horizon['horizon'].decompose()
            <Quantity 3628800. s>

            >>> from panoptes.utils import current_time
            >>> time_str = to_json({"current_time": current_time().datetime})
            >>> from_json(time_str)['current_time']         # doctest: +SKIP
            2019-04-08T06:43:28.232406

            >>> from astropy.time import Time
            >>> Time(from_json(time_str)['current_time'])   # doctest: +SKIP
            <Time object: scale='utc' format='isot' value=2019-04-08T06:43:28.232>

    Args:
        msg (`str`): The JSON string representation of the object.

    Returns:
        `dict`: The loaded object.
    """
    try:
        new_obj = _parse_all_objects(json.loads(msg))
    except json.decoder.JSONDecodeError as e:
        raise error.InvalidDeserialization(f'Error: {e!r} Message: {msg!r}')

    return new_obj


def to_yaml(obj, **kwargs):
    """Serialize a Python object to a YAML string.

    This will properly serialize the following:

        * `datetime.datetime`
        * `astropy.time.Time`
        * `astropy.units.Quantity`

    Examples:
        Also see the examples `from_yaml`.

        .. doctest::

            >>> import os
            >>> os.environ['POCSTIME'] = '1999-12-31 23:49:49'
            >>> from panoptes.utils import current_time
            >>> t0 = current_time()
            >>> t0
            <Time object: scale='utc' format='iso' value=1999-12-31 23:49:49.000>

            >>> to_yaml({'astropy time -> astropy time': t0})
            "astropy time -> astropy time: '1999-12-31T23:49:49.000'\\n"

            >>> to_yaml({'datetime -> astropy time': t0.datetime})
            "datetime -> astropy time: '1999-12-31T23:49:49.000'\\n"

            >>> # Can pass a `stream` parameter to save to file
            >>> with open('temp.yaml', 'w') as f:           # doctest: +SKIP
            ...     to_yaml({'my_object': 42}, stream=f)


    Args:
        obj (`object`): The object to be converted to be serialized.
        **kwargs: Arguments passed to `ruamel.yaml.dump`. See Examples.


    Returns:
        `str`: The YAML string representation of the object.
    """
    yaml = StringYAML()

    obj = _serialize_all_objects(deepcopy(obj))

    return yaml.dump(obj, **kwargs)


def from_yaml(msg, parse=True):
    """Convert a YAML string into a Python object.

    This is a thin-wrapper around `ruamel.YAML.load` that also parses the results
    looking for `astropy.units.Quantity` objects.

    Comments are preserved as long as the object remains YAML (lost on conversion
    to JSON, for example).

    See `from_json` for examples of astropy unit parsing.

    Examples:
        Note how comments in the YAML are preserved.

        .. doctest::

            >>> config_str = '''name: Generic PANOPTES Unit
            ... pan_id: PAN000
            ...
            ... location:
            ...   latitude: 19.54 deg
            ...   longitude: -155.58 deg
            ...   name: Mauna Loa Observatory  # Can be anything
            ... '''

            >>> config = from_yaml(config_str)
            >>> config['location']['latitude']
            <Quantity 19.54 deg>

            >>> yaml_config = to_yaml(config)
            >>> yaml_config                  # doctest: +SKIP
            ''' name: Generic PANOPTES Unit
            ... pan_id: PAN000  # CHANGE NAME
            ...
            ... location:
            ...   latitude: 19.54 deg
            ...   longitude: value: -155.58 deg
            ...   name: Mauna Loa Observatory  # Can be anything
            ... '''
            >>> yaml_config == config_str
            True

    Args:
        msg (`str`): The YAML string representation of the object.

    Returns:
        `collections.OrderedDict`: The ordered dict representing the YAML string, with appropriate
            object deserialization.
    """

    obj = YAML().load(msg)

    if parse:
        obj = _parse_all_objects(obj)

    return obj


def _parse_all_objects(obj):
    """Recursively parse the incoming object for astropy quantities.

    If `obj` is a dict with exactly two keys named `unit` and `value, then attempt
    to parse into a valid `astropy.unit.Quantity`. If fail, simply return object
    as is.

    Args:
        obj (`dict`): Object to check for quantities.

    Returns:
        `dict`: Same as `obj` but with objects converted to quantities.
    """
    if isinstance(obj, (dict, OrderedDict)):
        if 'value' in obj and 'unit' in obj:
            with suppress(ValueError):
                return obj['value'] * u.Unit(obj['unit'])

        for k, v in obj.items():
            obj[k] = _parse_all_objects(v)

    if isinstance(obj, bool):
        return bool(obj)

    # Try to turn into a time
    with suppress(KeyError, ValueError):
        if isinstance(Time(obj), Time):
            return Time(obj).datetime

    # Try to parse as quantity if certain type
    if isinstance(obj, str) and obj > '':
        with suppress(IndexError):
            units_string = obj.rsplit()[-1]  # Get the final word
            if units_string in ['m', 'deg', 's']:
                try:
                    quantity = u.Quantity(obj)
                    # If it ends up dimensionless just return obj.
                    if str(quantity.unit) == '':
                        return obj
                    else:
                        return quantity
                except Exception:
                    return obj

    return obj


def _serialize_all_objects(obj):
    for k, v in obj.items():
        # If it is a dict, send parse all its elements
        if isinstance(v, dict):
            obj[k] = _serialize_all_objects(v)
        else:
            obj[k] = _serialize_object(v, default=None)

    return obj


def _serialize_object(obj, default=None):
    # Astropy Quantity.
    if isinstance(obj, u.Quantity):
        return str(obj)

    # Astropy Time-like (including datetime).
    with suppress(ValueError):
        if isinstance(Time(obj), Time):
            return Time(obj).isot

    # Numpy array.
    if isinstance(obj, np.ndarray):
        return obj.tolist()

    # If we are given a default object type, e.g. str
    if default is not None:
        return default(obj)

    # Exceptions - if not a class object, then `issubclass` raises a `TypeError`,
    # so we ignore those and let the object pass through.
    with suppress(TypeError):
        if issubclass(obj, Exception):
            return str(obj)

    return obj



from datetime import datetime as dt
from datetime import timezone as tz

from google.auth.credentials import AnonymousCredentials
from google.cloud import firestore

from astropy.coordinates import SkyCoord
from astropy import units as u

import pandas as pd
import hvplot.pandas  # noqa

from dateutil.parser import parse as date_parse

from .. import listify
from ..logger import logger


def _get_firestore_client(project_id='panoptes-exp', database='(default)'):
    logger.debug(f'Getting new firestore client')
    firestore_client = firestore.Client(
        project=project_id,
        database=database,
        credentials=AnonymousCredentials()
    )
    return firestore_client


def get_metadata(image_id=None, sequence_id=None, fields=None, firestore_client=None):
    """Access PANOPTES data from the network.

    This function is capable of searching one type of object at a time, which is
    specified via the respective id parameter.

    #TODO(wtgee): Setup firestore emulator for testing. #179

    >>> from panoptes.utils.data import get_metadata
    >>> # Get image metadata as a DataFrame with one record.
    >>> image_id = 'PAN001_14d3bd_20181204T134406'
    >>> image_df = get_metadata(image_id=image_id)
    >>> type(image_df)
    <class 'pandas.core.frame.DataFrame'>
    >>> # This can easily be saved to csv or cast to dict.
    >>> image_metadata = image_df.to_dict(orient='record')[0]
    >>> image_metadata['status']
    'solved'
    >>> image_metadata['received_time']
    Timestamp('2020-04-17 09:16:20.371000+0000', tz='UTC')

    >>> # Get all image metadata for the observation.
    >>> sequence_id = 'PAN001_14d3bd_20181119T131353'
    >>> observation_df = get_metadata(sequence_id=sequence_id)
    >>> type(observation_df)
    <class 'pandas.core.frame.DataFrame'>
    >>> len(observation_df)
    40

    >>> # It's also possible to request certain fields
    >>> airmass_df = get_metadata(sequence_id=sequence_id, fields=['airmass'])
    >>> airmass_df.head()
        airmass
    0  1.161770
    1  1.166703
    2  1.172055
    3  1.177555
    4  1.183283

    Args:
        image_id (str|None): The id associated with an image.
        sequence_id (str|list|None): The list of sequence_ids associated with an observation.
        fields (list|None):  A list of fields to fetch from the database. If None,
            returns all fields.
        firestore_client (`google.cloud.firestore.Client`|None): The client instance
            to use. If `None` is provided (the default), then client will attempt
            to connect with default environmental credentials.

    Returns:

    """
    if firestore_client is None:
        firestore_client = _get_firestore_client()

    # Get a FITS image from the bucket.
    if image_id is not None:
        return get_image_metadata(image_id=image_id, fields=fields, firestore_client=firestore_client)

    # Get observation metadata from firestore.
    if sequence_id is not None:
        return get_observation_metadata(listify(sequence_id), fields=fields, firestore_client=firestore_client)


def get_image_metadata(image_id, fields=None, firestore_client=None):
    """Downloads the image at the given path.

    This function by default returns a `pandas.DataFrame` to be consistent with
    the `get_observation_metadata` function however that DataFrame should only contain
    a single row. Note that it will still be a DataFrame and not a `pandas.Series`.

    >>> from panoptes.utils.data import get_image_metadata
    >>> # Get image metadata as a DataFrame with one record.
    >>> image_id = 'PAN001_14d3bd_20181204T134406'
    >>> image_df = get_image_metadata(image_id=image_id, fields=['bucket_path'])
    >>> # Always includes the image_id and timestamp.
    >>> image_df.to_dict(orient='record')
    [{'bucket_path': 'PAN001/14d3bd/20181204T133735/20181204T134406.fits.fz',
      'image_id': 'PAN001_14d3bd_20181204T134406',
      'time': Timestamp('2018-12-04 13:44:06+0000', tz='UTC')}]
    >>> type(image_df)
    <class 'pandas.core.frame.DataFrame'>

    Args:
        image_id (str): The id for the given image.
        fields (list|None):  A list of fields to fetch from the database. If None,
            returns all fields.
        firestore_client (`google.cloud.firestore.Client`): An authenticated
            client instance. If None is provided then an anonymous connection will
            be made.

    Returns:
        `pandas.DataFrame`: DataFrame containing the image metadata.
    """
    if firestore_client is None:
        firestore_client = _get_firestore_client()

    logger.debug(f'Getting metadata for image={image_id}')

    # Get document reference.
    image_doc_ref = firestore_client.document(f'images/{image_id}')
    image_doc_snapshot = image_doc_ref.get()

    if image_doc_snapshot is None:
        logger.debug(f'No document found for image_id={image_id}')
        return

    # Get the actual image metadata.
    image_doc = image_doc_snapshot.to_dict()
    image_doc['image_id'] = image_doc_snapshot.id

    # Put document into dataframe.
    df = pd.DataFrame(image_doc, index=[0]).convert_dtypes()
    df = df.reindex(sorted(df.columns), axis=1)

    # Remove metadata metadata.
    # Remove fields if only certain fields requested.
    # TODO(wtgee) implement this filtering at the firestore level.
    if fields is not None:
        remove_cols = set(df.columns).difference(fields)
        # Don't remove the id or time.
        remove_cols.remove('image_id')
        remove_cols.remove('time')
        df.drop(columns=remove_cols, inplace=True)

    # TODO(wtgee) any data cleaning or preparation for images here.

    return df


def get_observation_metadata(sequence_ids, firestore_client=None, fields=None):
    """Get the metadata for given sequence_ids.

    Args:
        sequence_ids (list): A list of sequence_ids as strings.
        fields (list|None):  A list of fields to fetch from the database. If None,
            returns all fields.
        firestore_client (`google.cloud.firestore.Client`): An authenticated
            client instance. If None is provided then an anonymous connection will
            be made.

    Returns:
        `pandas.DataFrame`: DataFrame containing the observation metadata.
    """
    if firestore_client is None:
        firestore_client = _get_firestore_client()

    sequence_ids = listify(sequence_ids)

    observation_dfs = list()
    for sequence_id in sequence_ids:
        logger.debug(f'Getting images metadata for observation={sequence_id}')

        # Build query
        obs_query = firestore_client.collection('images').where('sequence_id', '==', sequence_id)

        # Fetch documents into a DataFrame.
        df = pd.DataFrame([dict(image_id=doc.id, **doc.to_dict()) for doc in obs_query.stream()])
        observation_dfs.append(df.convert_dtypes())

    if len(observation_dfs) == 0:
        logger.debug(f'No documents found for sequence_ids={sequence_ids}')
        return

    df = pd.concat(observation_dfs)
    df = df.reindex(sorted(df.columns), axis=1)
    df.sort_values(by=['time'], inplace=True)

    # TODO(wtgee) any data cleaning or preparation for observations here.

    # Remove fields if only certain fields requested.
    # TODO(wtgee) implement this filtering at the firestore level.
    if fields is not None:
        remove_cols = set(df.columns).difference(fields)
        df.drop(columns=remove_cols, inplace=True)

    return df


def search_observations(
        ra=None,
        dec=None,
        coords=None,
        radius=10,  # degrees
        start_date=None,
        end_date=None,
        unit_ids=None,
        status=None,
        min_num_images=1,
        fields=None,
        limit=5000,
        firestore_client=None,
):
    """Search PANOPTES observations.

    >>> from astropy.coordinates import SkyCoord
    >>> from panoptes.utils.data import search_observations
    >>> m42_coords = SkyCoord.from_name('M42')
    >>> search_results = search_observations(coords=m42_coords, radius=5, min_num_images=10, end_date='2020-04-27')
    >>> # The result is a DataFrame you can further work with.
    >>> search_results.groupby(['unit_id', 'field_name']).num_images.sum()
    unit_id  field_name
    PAN001   FlameNebula    436
             M42            422
    PAN006   M42             40
             Wasp 35         69
    Name: num_images, dtype: Int64
    >>> search_results.total_minutes_exptime.sum()
    1422.0

    Args:
        ra (float|None): The RA position in degrees of the center of search.
        dec (float|None): The Dec position in degrees of the center of the search.
        coords (`astropy.coordinates.SkyCoord`|None): A valid coordinate instance.
        radius (float): The search radius in degrees. Searches are currently done in
            a square box, so this is half the length of the side of the box.
        start_date (str|`datetime.datetime`|None): A valid datetime instance or `None` (default).
            If `None` then the beginning of 2018 is used as a start date.
        end_date (str|`datetime.datetime`|None): A valid datetime instance or `None` (default).
            If `None` then today is used.
        unit_ids (str|list|None): A str or list of strs of unit_ids to include.
            Default `None` will include all.
        status (str|list|None): A str or list of observation status to include.
            Default `None` will include all.
        min_num_images (int): Minimum number of images the observation should have, default 1.
        fields (str|list|None): A list of fields (columns) to include.
            Default `None` will include all.
        limit (int): The maximum number of firestore records to return, default 5000.
        firestore_client (`google.cloud.firestore.Client`|None): A firestore client instance.
            If default `None`, then the function will attempt to make an anonymous connection.

    Returns:
        `pandas.DataFrame`: A table with the matching observation results.
    """
    if firestore_client is None:
        firestore_client = _get_firestore_client()

    logger.debug(f'Setting up search params')

    if coords is None:
        coords = SkyCoord(ra=ra, dec=dec, unit='degree')

    # Setup defaults for search.
    if start_date is None:
        start_date = dt.strptime('2018-01-01', '%Y-%m-%d')
    elif isinstance(start_date, str):
        start_date = date_parse(start_date)

    if end_date is None:
        end_date = dt.now()
    elif isinstance(end_date, str):
        end_date = date_parse(end_date)

    ra_max = (coords.ra + (radius * u.degree)).value
    ra_min = (coords.ra - (radius * u.degree)).value
    dec_max = (coords.dec + (radius * u.degree)).value
    dec_min = (coords.dec - (radius * u.degree)).value

    logger.debug(f'Searching for observations')
    obs_query = firestore_client.collection('observations') \
        .where('dec', '>=', dec_min) \
        .where('dec', '<=', dec_max) \
        .limit(limit)

    # Fetch documents.
    docs = [dict(sequence_id=doc.id, **doc.to_dict()) for doc in obs_query.stream()]

    if len(docs) == 0:
        logger.debug(f'No documents found for collections query')
        return

    # Put documents into a DataFrame.
    df = pd.DataFrame(docs).convert_dtypes()
    logger.debug(f'Found {len(df)} observations')

    # Perform filtering on other fields here.
    logger.debug(f'Filtering observations')
    df.query(
        f'ra >= {ra_min} and ra <= {ra_max}'
        ' and '
        f'time >= "{start_date.astimezone(tz.utc)}" and time <= "{end_date.astimezone(tz.utc)}"'
        ' and '
        f'num_images >= {min_num_images}'
        ,
        inplace=True
    )
    if unit_ids is not None and unit_ids != 'The Whole World! 🌎':
        df.query(f'unit_id in {listify(unit_ids)}', inplace=True)

    if status is not None:
        df.query(f'status in {listify(status)}', inplace=True)

    logger.debug(f'Found {len(df)} observations after filtering')

    df = df.reindex(sorted(df.columns), axis=1)
    df.sort_values(by=['time'], inplace=True)

    # TODO(wtgee) any data cleaning or preparation for observations here.

    # Remove fields if only certain fields requested.
    # TODO(wtgee) implement this filtering at the firestore level.
    if fields is not None:
        remove_cols = set(df.columns).difference(listify(fields))
        df.drop(columns=remove_cols, inplace=True)

    return df



#!/usr/bin/env python
# Licensed under an MIT style license - see LICENSE.txt

from configparser import ConfigParser
from setuptools import setup, find_namespace_packages

import versioneer


# Get some values from the setup.cfg
conf = ConfigParser()
conf.read(['setup.cfg'])
metadata = dict(conf.items('metadata'))

AUTHOR = metadata.get('author', '')
AUTHOR_EMAIL = metadata.get('author_email', '')
DESCRIPTION = metadata.get('description', '')
KEYWORDS = metadata.get('keywords', 'Project PANOPTES')
LICENSE = metadata.get('license', 'unknown')
LONG_DESCRIPTION = metadata.get('long_description', '')
NAME = metadata.get('name', 'panoptes-utils')
PACKAGENAME = metadata.get('package_name', 'packagename')
URL = metadata.get('url', 'https://projectpanoptes.org')

modules = {
    'required': [
        'astroplan>=0.6',
        'astropy>=4.0.0',
        'coverage',  # testing
        'Flask',
        'fastparquet',
        'google-cloud-bigquery[pandas]',
        'google-cloud-firestore',
        'google-cloud-storage',
        'holoviews',
        'hvplot',
        'loguru',
        'matplotlib>=3.1.3',
        'mocket',  # testing
        'numpy',
        'pandas',
        'photutils',
        'Pillow',
        'pycodestyle',  # testing
        'pyserial',
        'pytest',  # testing
        'pytest-cov',  # testing
        'pytest-remotedata>=0.3.1',  # testing
        'python-dateutil',
        'PyYAML',
        'pyzmq',
        'requests',  # social
        'ruamel.yaml>=0.15',
        'scalpl',
        'scipy',
        'tweepy',  # social
        'versioneer',
    ],
}


setup(name=NAME,
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      url=URL,
      keywords=KEYWORDS,
      python_requires='>=3.6',
      setup_requires=['pytest-runner'],
      tests_require=modules['required'],
      scripts=[
          'bin/cr2-to-jpg',
          'bin/panoptes-config-server',
          'bin/panoptes-messaging-hub',
          'bin/panoptes-solve-field',
          'bin/panoptes-develop',
      ],
      install_requires=modules['required'],
      packages=find_namespace_packages(include=['panoptes.utils.*'],
                                       exclude=['tests', 'test_*']),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Operating System :: POSIX',
          'Programming Language :: C',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3 :: Only',
          'Topic :: Scientific/Engineering :: Astronomy',
          'Topic :: Scientific/Engineering :: Physics',
      ],
      zip_safe=False
      )



import ctypes
import ctypes.util
from panoptes.utils.logger import logger

from astropy.utils import resolve_name
from . import error


def load_c_library(name, path=None, mode=ctypes.DEFAULT_MODE, **kwargs):
    """Utility function to load a shared/dynamically linked library (.so/.dylib/.dll).

    The name and location of the shared library can be manually specified with the library_path
    argument, otherwise the ctypes.util.find_library function will be used to try to locate based
    on library_name.

    Args:
        name (str): name of the library (without 'lib' prefix or any suffixes, e.g. 'fli').
        path (str, optional): path to the library e.g. '/usr/local/lib/libfli.so'.
        mode (int, optional): mode in which to load the library, see dlopen(3) man page for
            details. Should be one of ctypes.RTLD_GLOBAL, ctypes.RTLD_LOCAL, or
            ctypes.DEFAULT_MODE. Default is ctypes.DEFAULT_MODE.

    Returns:
        ctypes.CDLL

    Raises:
        pocs.utils.error.NotFound: raised if library_path not given & find_library fails to
            locate the library.
        OSError: raises if the ctypes.CDLL loader cannot load the library.
    """
    if mode is None:
        # Interpret a value of None as the default.
        mode = ctypes.DEFAULT_MODE
    # Open library
    logger.debug(f"Opening {name} library")
    if not path:
        path = ctypes.util.find_library(name)
        if not path:
            raise error.NotFound(f"Cound not find {name} library!")
    # This CDLL loader will raise OSError if the library could not be loaded
    return ctypes.CDLL(path, mode=mode)


def load_module(module_name):
    """Dynamically load a module.

    >>> from panoptes.utils.library import load_module
    >>> error = load_module('panoptes.utils.error')
    >>> error.__name__
    'panoptes.utils.error'
    >>> error.__package__
    'panoptes.utils'

    Args:
        module_name (str): Name of module to import.

    Returns:
        module: an imported module name

    Raises:
        error.NotFound: If module cannot be imported.
    """
    try:
        module = resolve_name(module_name)
    except ImportError:
        raise error.NotFound(msg=module_name)

    return module



# This module enables a test to provide a handler for "hooked://..." urls
# passed into serial.serial_for_url. To do so, set the value of
# serial_class_for_url from your test to a function with the same API as
# ExampleSerialClassForUrl. Or assign your class to Serial.

from panoptes.utils.tests.serial_handlers import NoOpSerial


def ExampleSerialClassForUrl(url):
    """Implementation of serial_class_for_url called by serial.serial_for_url.

    Returns the url, possibly modified, and a factory function to be called to
    create an instance of a SerialBase sub-class (or at least behaves like it).
    You can return a class as that factory function, as calling a class creates
    an instance of that class.

    serial.serial_for_url will call that factory function with None as the
    port parameter (the first), and after creating the instance will assign
    the url to the port property of the instance.

    Returns:
        A tuple (url, factory).
    """
    return url, Serial


# Assign to this global variable from a test to override this default behavior.
serial_class_for_url = ExampleSerialClassForUrl

# Or assign your own class to this global variable.
Serial = NoOpSerial



import os
import subprocess
import shutil

from dateutil import parser as date_parser
from json import loads
from warnings import warn

import numpy as np

from astropy.io import fits

from .. import error
from ..logger import logger
from ..images import fits as fits_utils


def cr2_to_fits(
        cr2_fname,
        fits_fname=None,
        overwrite=False,
        headers={},
        fits_headers={},
        remove_cr2=False,
        **kwargs):  # pragma: no cover
    """Convert a CR2 file to FITS

    This is a convenience function that first converts the CR2 to PGM via ~cr2_to_pgm.
    Also adds keyword headers to the FITS file.

    Note:
        The intermediate PGM file is automatically removed

    Arguments:
        cr2_fname (str): Name of the CR2 file to be converted.
        fits_fname (str, optional): Name of the FITS file to output. Default is `None`, in which
            case the `cr2_fname` is used as the base.
        overwrite (bool, optional): Overwrite existing FITS, default False.
        headers (dict, optional): Header data added to the FITS file.
        fits_headers (dict, optional): Header data added to the FITS file without filtering.
        remove_cr2 (bool, optional): If CR2 should be removed after processing, default False.
        **kwargs: Description

    Returns:
        str: The full path to the generated FITS file.

    """
    if fits_fname is None:
        fits_fname = cr2_fname.replace('.cr2', '.fits')

    if not os.path.exists(fits_fname) or overwrite:
        logger.debug("Converting CR2 to PGM: {}".format(cr2_fname))

        # Convert the CR2 to a PGM file then delete PGM
        pgm = read_pgm(cr2_to_pgm(cr2_fname), remove_after=True)

        # Add the EXIF information from the CR2 file
        exif = read_exif(cr2_fname)

        # Set the PGM as the primary data for the FITS file
        hdu = fits.PrimaryHDU(pgm)

        obs_date = date_parser.parse(
            exif.get('DateTimeOriginal', '').replace(':', '-', 2)).isoformat()

        # Set some default headers
        hdu.header.set('FILTER', 'RGGB')
        hdu.header.set('ISO', exif.get('ISO', ''))
        hdu.header.set('EXPTIME', exif.get('ExposureTime', 'Seconds'))
        hdu.header.set('CAMTEMP', exif.get('CameraTemperature', ''), 'Celsius - From CR2')
        hdu.header.set('CIRCCONF', exif.get('CircleOfConfusion', ''), 'From CR2')
        hdu.header.set('COLORTMP', exif.get('ColorTempMeasured', ''), 'From CR2')
        hdu.header.set('FILENAME', exif.get('FileName', ''), 'From CR2')
        hdu.header.set('INTSN', exif.get('InternalSerialNumber', ''), 'From CR2')
        hdu.header.set('CAMSN', exif.get('SerialNumber', ''), 'From CR2')
        hdu.header.set('MEASEV', exif.get('MeasuredEV', ''), 'From CR2')
        hdu.header.set('MEASEV2', exif.get('MeasuredEV2', ''), 'From CR2')
        hdu.header.set('MEASRGGB', exif.get('MeasuredRGGB', ''), 'From CR2')
        hdu.header.set('WHTLVLN', exif.get('NormalWhiteLevel', ''), 'From CR2')
        hdu.header.set('WHTLVLS', exif.get('SpecularWhiteLevel', ''), 'From CR2')
        hdu.header.set('REDBAL', exif.get('RedBalance', ''), 'From CR2')
        hdu.header.set('BLUEBAL', exif.get('BlueBalance', ''), 'From CR2')
        hdu.header.set('WBRGGB', exif.get('WB RGGBLevelAsShot', ''), 'From CR2')
        hdu.header.set('DATE-OBS', obs_date)

        for key, value in fits_headers.items():
            try:
                hdu.header.set(key.upper()[0: 8], value)
            except Exception:
                pass

        try:
            logger.debug("Saving fits file to: {}".format(fits_fname))

            hdu.writeto(fits_fname, output_verify='silentfix', overwrite=overwrite)
        except Exception as e:
            warn("Problem writing FITS file: {}".format(e))
        else:
            if remove_cr2:
                os.unlink(cr2_fname)

        fits_utils.update_observation_headers(fits_fname, headers)

    return fits_fname


def cr2_to_pgm(
        cr2_fname,
        pgm_fname=None,
        overwrite=True, *args,
        **kwargs):  # pragma: no cover
    """ Convert CR2 file to PGM

    Converts a raw Canon CR2 file to a netpbm PGM file via `dcraw`. Assumes
    `dcraw` is installed on the system

    Note:
        This is a blocking call

    Arguments:
        cr2_fname {str} -- Name of CR2 file to convert
        **kwargs {dict} -- Additional keywords to pass to script

    Keyword Arguments:
        pgm_fname {str} -- Name of PGM file to output, if None (default) then
                           use same name as CR2 (default: {None})
        dcraw {str} -- Path to installed `dcraw` (default: {'dcraw'})
        overwrite {bool} -- A bool indicating if existing PGM should be overwritten
                         (default: {True})

    Returns:
        str -- Filename of PGM that was created

    """
    dcraw = shutil.which('dcraw')
    if dcraw is None:
        raise error.InvalidCommand('dcraw not found')

    if pgm_fname is None:
        pgm_fname = cr2_fname.replace('.cr2', '.pgm')

    if os.path.exists(pgm_fname) and not overwrite:
        logger.warning(f"PGM file exists, returning existing file: {pgm_fname}")
    else:
        try:
            # Build the command for this file
            command = '{} -t 0 -D -4 {}'.format(dcraw, cr2_fname)
            cmd_list = command.split()
            logger.debug("PGM Conversion command: \n {}".format(cmd_list))

            # Run the command
            if subprocess.check_call(cmd_list) == 0:
                logger.debug("PGM Conversion command successful")

        except subprocess.CalledProcessError as err:
            raise error.InvalidSystemCommand(msg="File: {} \n err: {}".format(cr2_fname, err))

    return pgm_fname


def read_exif(fname, exiftool='exiftool'):  # pragma: no cover
    """ Read the EXIF information

    Gets the EXIF information using exiftool

    Note:
        Assumes the `exiftool` is installed

    Args:
        fname {str} -- Name of file (CR2) to read

    Keyword Args:
        exiftool {str} -- Location of exiftool (default: {'/usr/bin/exiftool'})

    Returns:
        dict -- Dictonary of EXIF information

    """
    assert os.path.exists(fname), warn("File does not exist: {}".format(fname))
    exif = {}

    try:
        # Build the command for this file
        command = '{} -j {}'.format(exiftool, fname)
        cmd_list = command.split()

        # Run the command
        exif = loads(subprocess.check_output(cmd_list).decode('utf-8'))
    except subprocess.CalledProcessError as err:
        raise error.InvalidSystemCommand(
            msg="File: {} \n err: {}".format(fname, err))

    return exif[0]


def read_pgm(fname, byteorder='>', remove_after=False):  # pragma: no cover
    """Return image data from a raw PGM file as numpy array.

    Note:
        Format Spec: http://netpbm.sourceforge.net/doc/pgm.html
        Source: http://stackoverflow.com/questions/7368739/numpy-and-16-bit-pgm

    Note:
        This is correctly processed as a Big endian even though the CR2 itself
        marks it as a Little endian. See the notes in Source page above as well
        as the comment about significant bit in the Format Spec

    Args:
        fname(str):         Filename of PGM to be converted
        byteorder(str):     Big endian
        remove_after(bool): Delete fname file after reading, defaults to False.
        overwrite(bool):      overwrite existing PGM or not, defaults to True

    Returns:
        numpy.array:        The raw data from the PGMx

    """

    with open(fname, 'rb') as f:
        buffer = f.read()

    # We know our header info is 19 chars long
    header_offset = 19

    img_type, img_size, img_max_value, _ = buffer[
        0:header_offset].decode().split('\n')

    assert img_type == 'P5', warn("No a PGM file")

    # Get the width and height (as strings)
    width, height = img_size.split(' ')

    data = np.flipud(np.frombuffer(buffer[header_offset:],
                                   dtype=byteorder + 'u2',
                                   ).reshape((int(height), int(width))))

    if remove_after:
        os.remove(fname)

    return data



import os
import time
from datetime import timezone as tz

from astropy import units as u
from astropy.time import Time

from .logger import logger


def current_time(flatten=False, datetime=False, pretty=False):
    """ Convenience method to return the "current" time according to the system.

    Note:
        If the ``$POCSTIME`` environment variable is set then this will return
        the time given in the variable. This is used for setting specific times
        during testing. After checking the value of POCSTIME the environment
        variable will also be incremented by one second so that subsequent
        calls to this function will generate monotonically increasing times.

        **Operation of POCS from $POCS/bin/pocs_shell will clear the POCSTIME variable.**

        .. doctest::

            >>> os.environ['POCSTIME'] = '1999-12-31 23:59:59'
            >>> party_time = current_time(pretty=True)
            >>> party_time
            '1999-12-31 23:59:59'

            # Next call is one second later when using $POCSTIME.
            >>> y2k = current_time(pretty=True)
            >>> y2k
            '2000-01-01 00:00:00'


    Note:
        The time returned from this function is **not** timezone aware. All times
        are UTC.

    .. doctest::

        >>> from panoptes.utils import current_time
        >>> current_time()                # doctest: +SKIP
        <Time object: scale='utc' format='datetime' value=2018-10-07 22:29:03.009873>

        >>> current_time(datetime=True)   # doctest: +SKIP
        datetime.datetime(2018, 10, 7, 22, 29, 26, 594368)

        >>> current_time(pretty=True)     # doctest: +SKIP
        '2018-10-07 22:29:51'


    Returns:
        astropy.time.Time: Object representing now.
    """

    pocs_time = os.getenv('POCSTIME')

    if pocs_time is not None and pocs_time > '':
        _time = Time(pocs_time)
        # Increment POCSTIME
        os.environ['POCSTIME'] = (_time + 1 * u.second).isot
    else:
        _time = Time.now()

    if flatten:
        _time = flatten_time(_time)

    if pretty:
        _time = _time.isot.split('.')[0].replace('T', ' ')

    if datetime:
        # Add UTC timezone
        _time = _time.to_datetime(timezone=tz.utc)

    return _time


def flatten_time(t):
    """Given an astropy time, flatten to have no extra chars besides integers.

    .. doctest::

        >>> from astropy.time import Time
        >>> from panoptes.utils import flatten_time
        >>> t0 = Time('1999-12-31 23:59:59')
        >>> t0.isot
        '1999-12-31T23:59:59.000'

        >>> flatten_time(t0)
        '19991231T235959'

    Args:
        t (astropy.time.Time): The time to be flattened.

    Returns:
        str: The flattened string representation of the time.
    """
    return t.isot.replace('-', '').replace(':', '').split('.')[0]


# This is a streamlined variant of PySerial's serialutil.Timeout.
class CountdownTimer(object):
    """Simple timer object for tracking whether a time duration has elapsed.


    Args:
        duration (int or float or astropy.units.Quantity): Amount of time to before time expires.
            May be numeric seconds or an Astropy time duration (e.g. 1 * u.minute).
    """

    def __init__(self, duration):
        if isinstance(duration, u.Quantity):
            duration = duration.to(u.second).value
        elif not isinstance(duration, (int, float)):
            raise ValueError(f'duration ({duration}) is not a supported type: {type(duration)}')

        #: bool: True IFF the duration is zero.
        assert duration >= 0, "Duration must be non-negative."
        self.is_non_blocking = (duration == 0)

        self.duration = float(duration)
        self.restart()

    def __str__(self):
        is_blocking = ''
        if self.is_non_blocking is False:
            is_blocking = '(blocking)'
        is_expired = ''
        if self.expired():
            is_expired = 'EXPIRED '
        return f'{is_expired}Timer {is_blocking} {self.time_left():.02f}/{self.duration:.02f}'

    def expired(self):
        """Return a boolean, telling if the timeout has expired.

        Returns:
            bool: If timer has expired.
        """
        return self.time_left() <= 0

    def time_left(self):
        """Return how many seconds are left until the timeout expires.

        Returns:
            int: Number of seconds remaining in timer, zero if ``is_non_blocking=True``.
        """
        if self.is_non_blocking:
            return 0
        else:
            delta = self.target_time - time.monotonic()
            if delta > self.duration:
                # clock jumped, recalculate
                self.restart()
                return self.duration
            else:
                return max(0.0, delta)

    def restart(self):
        """Restart the timed duration."""
        self.target_time = time.monotonic() + self.duration
        logger.debug(f'Restarting {self}')

    def sleep(self, max_sleep=None):
        """Sleep until the timer expires, or for max_sleep, whichever is sooner.

        Args:
            max_sleep: Number of seconds to wait for, or None.
        Returns:
            True if slept for less than time_left(), False otherwise.
        """
        # Sleep for remaining time by default.
        remaining = self.time_left()
        if not remaining:
            return False
        sleep_time = remaining

        # Sleep only for max time if requested.
        if max_sleep and max_sleep < remaining:
            assert max_sleep > 0
            sleep_time = max_sleep

        logger.debug(f'Sleeping for {sleep_time:.02f} seconds')
        time.sleep(sleep_time)

        return sleep_time < remaining



import os
import pytest

from astropy.io import fits

from panoptes.utils.images import mask_saturated
from panoptes.utils.images import focus as focus_utils


def test_vollath_f4(data_dir):
    data = fits.getdata(os.path.join(data_dir, 'unsolved.fits'))
    data = mask_saturated(data)
    assert focus_utils.vollath_F4(data) == pytest.approx(14667.207897717599)
    assert focus_utils.vollath_F4(data, axis='Y') == pytest.approx(14380.343807477504)
    assert focus_utils.vollath_F4(data, axis='X') == pytest.approx(14954.071987957694)
    with pytest.raises(ValueError):
        focus_utils.vollath_F4(data, axis='Z')


def test_focus_metric_default(data_dir):
    data = fits.getdata(os.path.join(data_dir, 'unsolved.fits'))
    data = mask_saturated(data)
    assert focus_utils.focus_metric(data) == pytest.approx(14667.207897717599)
    assert focus_utils.focus_metric(data, axis='Y') == pytest.approx(14380.343807477504)
    assert focus_utils.focus_metric(data, axis='X') == pytest.approx(14954.071987957694)
    with pytest.raises(ValueError):
        focus_utils.focus_metric(data, axis='Z')


def test_focus_metric_vollath(data_dir):
    data = fits.getdata(os.path.join(data_dir, 'unsolved.fits'))
    data = mask_saturated(data)
    assert focus_utils.focus_metric(
        data, merit_function='vollath_F4') == pytest.approx(14667.207897717599)
    assert focus_utils.focus_metric(
        data,
        merit_function='vollath_F4',
        axis='Y') == pytest.approx(14380.343807477504)
    assert focus_utils.focus_metric(
        data,
        merit_function='vollath_F4',
        axis='X') == pytest.approx(14954.071987957694)
    with pytest.raises(ValueError):
        focus_utils.focus_metric(data, merit_function='vollath_F4', axis='Z')


def test_focus_metric_bad_string(data_dir):
    data = fits.getdata(os.path.join(data_dir, 'unsolved.fits'))
    data = mask_saturated(data)
    with pytest.raises(KeyError):
        focus_utils.focus_metric(data, merit_function='NOTAMERITFUNCTION')



"""Provides a simple simulator for telemetry_board.ino or camera_board.ino.

We use the pragma "no cover" in several places that happen to never be
reached or that would only be reached if the code was called directly,
i.e. not in the way it is intended to be used.
"""

import copy
import datetime
import queue
import random
from serial import serialutil
import threading
import time
import urllib

from panoptes.utils.logger import logger
from panoptes.utils.tests import serial_handlers
from panoptes.utils.serializers import to_json
from panoptes.utils.serializers import from_json


def _drain_queue(q):
    cmd = None
    while not q.empty():
        cmd = q.get_nowait()
    return cmd  # Present just for debugging.


class ArduinoSimulator:
    """Simulates the serial behavior of the PANOPTES Arduino sketches.

    The RS-232 connection is simulated with an input and output queue of bytes. This class provides
    a run function which can be called from a Thread to execute. Every two seconds while running it
    will generate another json output line, and then send that to the json_queue in small chunks
    at a rate similar to 9600 baud, the rate used by our Arduino sketches.
    """

    def __init__(self, message, relay_queue, json_queue, chunk_size, stop):
        """
        Args:
            message: The message to be sent (millis and report_num will be added).
            relay_queue: The queue.Queue instance from which relay command
                bytes are read and acted upon. Elements are of type bytes.
            json_queue: The queue.Queue instance to which json messages
                (serialized to bytes) are written at ~9600 baud. Elements
                are of type bytes (i.e. each element is a sequence of bytes of
                length up to chunk_size).
            chunk_size: The number of bytes to write to json_queue at a time.
            stop: a threading.Event which is checked to see if run should stop executing.
        """
        self.message = copy.deepcopy(message)
        self.logger = logger
        self.logger.critical(f'message: {message}')
        self.relay_queue = relay_queue
        self.json_queue = json_queue
        self.stop = stop
        # Time between producing messages.
        self.message_delta = datetime.timedelta(seconds=2)
        self.next_message_time = None
        # Size of a chunk of bytes.
        self.chunk_size = chunk_size
        # Interval between outputing chunks of bytes.
        chunks_per_second = 1000.0 / self.chunk_size
        chunk_interval = 1.0 / chunks_per_second
        self.logger.debug('chunks_per_second={}   chunk_interval={}', chunks_per_second,
                          chunk_interval)
        self.chunk_delta = datetime.timedelta(seconds=chunk_interval)
        self.next_chunk_time = None
        self.pending_json_bytes = bytearray()
        self.pending_relay_bytes = bytearray()
        self.command_lines = []
        self.start_time = datetime.datetime.now()
        self.report_num = 0
        self.logger.info('ArduinoSimulator created')

    def __del__(self):
        if not self.stop.is_set():  # pragma: no cover
            self.logger.critical('ArduinoSimulator.__del__ stop is NOT set')

    def run(self):
        """Produce messages periodically and emit their bytes at a limited rate."""
        self.logger.info('ArduinoSimulator.run ENTER')
        # Produce a message right away, but remove a random number of bytes at the start to reflect
        # what happens when we connect at a random time to the Arduino.
        now = datetime.datetime.now()
        self.next_chunk_time = now
        self.next_message_time = now + self.message_delta
        b = self.generate_next_message_bytes(now)
        cut = random.randrange(len(b))
        if cut > 0:
            self.logger.info('Cutting off the leading {} bytes of the first message', cut)
            b = b[cut:]
        self.pending_json_bytes.extend(b)
        # Now two interleaved loops:
        # 1) Generate messages every self.message_delta
        # 2) Emit a chunk of bytes from pending_json_bytes every self.chunk_delta.
        # Clearly we need to emit all the bytes from pending_json_bytes at least
        # as fast as we append new messages to it, else we'll have a problem
        # (i.e. the simulated baud rate will be too slow for the output rate).
        while True:
            if self.stop.is_set():
                self.logger.info('Returning from ArduinoSimulator.run EXIT')
                return
            now = datetime.datetime.now()
            if now >= self.next_chunk_time:
                self.output_next_chunk(now)
            if now >= self.next_message_time:
                self.generate_next_message(now)
            if self.pending_json_bytes and self.next_chunk_time < self.next_message_time:
                next_time = self.next_chunk_time
            else:
                next_time = self.next_message_time
            self.read_relay_queue_until(next_time)

    def handle_pending_relay_bytes(self):
        """Process complete relay commands."""
        newline = b'\n'
        while True:
            index = self.pending_relay_bytes.find(newline)
            if index < 0:
                break
            line = str(self.pending_relay_bytes[0:index], 'ascii')
            self.logger.info(f'Received command: {line}')
            del self.pending_relay_bytes[0:index + 1]
            self.command_lines.append(line)
        if self.pending_relay_bytes:
            self.logger.info(f'Accumulated {len(self.pending_relay_bytes)} bytes.')

    def read_relay_queue_until(self, next_time):
        """Read and process relay queue bytes until time for the next action."""
        while True:
            now = datetime.datetime.now()
            if now >= next_time:
                # Already reached the time for the next main loop event,
                # so return to repeat the main loop.
                return
            remaining = (next_time - now).total_seconds()
            assert remaining > 0
            self.logger.info('ArduinoSimulator.read_relay_queue_until remaining={}', remaining)
            try:
                b = self.relay_queue.get(block=True, timeout=remaining)
                assert isinstance(b, (bytes, bytearray))
                self.pending_relay_bytes.extend(b)
                self.handle_pending_relay_bytes()
                # Fake a baud rate for reading by waiting based on the
                # number of bytes we just read.
                time.sleep(1.0 / 1000 * len(b))
            except queue.Empty:
                # Not returning here so that the return above is will be
                # hit every time this method executes.
                pass

    def output_next_chunk(self, now):
        """Output one chunk of pending json bytes."""
        self.next_chunk_time = now + self.chunk_delta
        if len(self.pending_json_bytes) == 0:
            return
        last = min(self.chunk_size, len(self.pending_json_bytes))
        chunk = bytes(self.pending_json_bytes[0:last])
        del self.pending_json_bytes[0:last]
        if self.json_queue.full():
            self.logger.info('Dropping chunk because the queue is full')
            return
        self.json_queue.put_nowait(chunk)
        self.logger.debug('output_next_chunk -> {}', chunk)

    def generate_next_message(self, now):
        """Append the next message to the pending bytearray and scheduled the next message."""
        b = self.generate_next_message_bytes(now)
        self.pending_json_bytes.extend(b)
        self.next_message_time = datetime.datetime.now() + self.message_delta

    def generate_next_message_bytes(self, now):
        """Generate the next message (report) from the simulated Arduino."""
        # Not worrying here about emulating the 32-bit nature of millis (wraps in 49 days)
        elapsed = int((now - self.start_time).total_seconds() * 1000)
        self.report_num += 1
        self.message['millis'] = elapsed
        self.message['report_num'] = self.report_num
        if self.command_lines:
            self.message['commands'] = self.command_lines
            self.command_lines = []

        s = to_json(self.message) + '\r\n'
        if 'commands' in self.message:
            del self.message['commands']

        self.logger.debug('generate_next_message -> {!r}', s)
        b = s.encode(encoding='ascii')
        return b


class FakeArduinoSerialHandler(serial_handlers.NoOpSerial):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logger
        self.simulator_thread = None
        self.relay_queue = queue.Queue(maxsize=1)
        self.json_queue = queue.Queue(maxsize=1)
        self.json_bytes = bytearray()
        self.stop = threading.Event()
        self.stop.set()
        self.device_simulator = None

    def __del__(self):
        if self.simulator_thread:  # pragma: no cover
            self.logger.critical('ArduinoSimulator.__del__ simulator_thread is still present')
            self.stop.set()
            self.simulator_thread.join(timeout=3.0)

    def open(self):
        """Open port.

        Raises:
            SerialException if the port cannot be opened.
        """
        if not self.is_open:
            self.is_open = True
            self._reconfigure_port()

    def close(self):
        """Close port immediately."""
        self.is_open = False
        self._reconfigure_port()

    @property
    def in_waiting(self):
        """The number of input bytes available to read immediately."""
        if not self.is_open:
            raise serialutil.portNotOpenError
        # Not an accurate count because the elements of self.json_queue are arrays, not individual
        # bytes.
        return len(self.json_bytes) + self.json_queue.qsize()

    def reset_input_buffer(self):
        """Flush input buffer, discarding all it’s contents."""
        self.json_bytes.clear()
        _drain_queue(self.json_queue)

    def read(self, size=1):
        """Read size bytes.

        If a timeout is set it may return fewer characters than requested.
        With no timeout it will block until the requested number of bytes
        is read.

        Args:
            size: Number of bytes to read.

        Returns:
            Bytes read from the port, of type 'bytes'.
        """
        if not self.is_open:
            raise serialutil.portNotOpenError

        # Not checking if the config is OK, so will try to read from a possibly
        # empty queue if using the wrong baudrate, etc. This is deliberate.

        response = bytearray()
        timeout_obj = serialutil.Timeout(self.timeout)
        while True:
            b = self._read1(timeout_obj)
            if b:
                response.extend(b)
                if size is not None and len(response) >= size:
                    break
            else:  # pragma: no cover
                # The timeout expired while in _read1.
                break
            if timeout_obj.expired():  # pragma: no cover
                break
        response = bytes(response)
        return response

    def readline(self):
        """Read and return one line from the simulator.

        This override exists just to support logging of the line.
        """
        line = super().readline()
        self.logger.debug('FakeArduinoSerialHandler.readline -> {!r}', line)
        return line

    @property
    def out_waiting(self):
        """The number of bytes in the output buffer."""
        if not self.is_open:
            raise serialutil.portNotOpenError
        # Not an accurate count because the elements of self.relay_queue are arrays, not individual
        # bytes.
        return self.relay_queue.qsize()

    def reset_output_buffer(self):
        """Clear output buffer.

        Aborts the current output, discarding all that is in the output buffer.
        """
        if not self.is_open:
            raise serialutil.portNotOpenError
        _drain_queue(self.relay_queue)

    def flush(self):
        """Write the buffered data to the output device.

        We interpret that here as waiting until the simulator has taken all of the
        entries from the queue.
        """
        if not self.is_open:
            raise serialutil.portNotOpenError
        while not self.relay_queue.empty():
            time.sleep(0.01)

    def write(self, data):
        """Write the bytes data to the port.

        Args:
            data: The data to write (bytes or bytearray instance).

        Returns:
            Number of bytes written.

        Raises:
            SerialTimeoutException: In case a write timeout is configured for
                the port and the time is exceeded.
        """
        if not isinstance(data, (bytes, bytearray)):
            raise ValueError('write takes bytes')  # pragma: no cover
        data = bytes(data)  # Make sure it can't change.
        self.logger.info('FakeArduinoSerialHandler.write({!r})', data)
        try:
            for n in range(len(data)):
                one_byte = data[n:n + 1]
                self.relay_queue.put(one_byte, block=True, timeout=self.write_timeout)
            return len(data)
        except queue.Full:  # pragma: no cover
            # This exception is "lossy" in that the caller can't tell how much was written.
            raise serialutil.writeTimeoutError

    # --------------------------------------------------------------------------

    @property
    def is_config_ok(self):
        """Does the caller ask for the correct serial device config?"""
        # The default Arduino data, parity and stop bits are: 8 data bits, no parity, one stop bit.
        v = (self.baudrate == 9600 and self.bytesize == serialutil.EIGHTBITS and
             self.parity == serialutil.PARITY_NONE and not self.rtscts and not self.dsrdtr)

        # All existing tests ensure the config is OK, so we never log here.
        if not v:  # pragma: no cover
            self.logger.critical('Serial config is not OK: {!r}', (self.get_settings(), ))

        return v

    def _read1(self, timeout_obj):
        """Read 1 byte of input, of type bytes."""

        # _read1 is currently called only from read(), which checks that the
        # serial device is open, so is_open is always true.
        if not self.is_open:  # pragma: no cover
            raise serialutil.portNotOpenError

        if not self.json_bytes:
            try:
                entry = self.json_queue.get(block=True, timeout=timeout_obj.time_left())
                assert isinstance(entry, bytes)
                self.json_bytes.extend(entry)
            except queue.Empty:
                return None

        # Unless something has gone wrong, json_bytes is always non-empty here.
        if not self.json_bytes:  # pragma: no cover
            return None

        c = bytes(self.json_bytes[0:1])
        del self.json_bytes[0:1]
        return c

    # --------------------------------------------------------------------------
    # There are a number of methods called by SerialBase that need to be
    # implemented by sub-classes, assuming their calls haven't been blocked
    # by replacing the calling methods/properties. These are no-op
    # implementations.

    def _reconfigure_port(self):
        """Reconfigure the open port after a property has been changed.

        If you need to know which property has been changed, override the
        setter for the appropriate properties.
        """
        need_thread = self.is_open and self.is_config_ok
        if need_thread and not self.simulator_thread:
            _drain_queue(self.relay_queue)
            _drain_queue(self.json_queue)
            self.json_bytes.clear()
            self.stop.clear()
            params = self._params_from_url(self.portstr)
            self._create_simulator(params)
            self.simulator_thread = threading.Thread(
                name='Device Simulator', target=lambda: self.device_simulator.run(), daemon=True)
            self.simulator_thread.start()
        elif self.simulator_thread and not need_thread:
            self.stop.set()
            self.simulator_thread.join(timeout=30.0)
            if self.simulator_thread.is_alive():
                # Not a SerialException, but a test infrastructure error.
                raise Exception(self.simulator_thread.name + ' thread did not stop!')  # pragma: no cover
            self.simulator_thread = None
            self.device_simulator = None
            _drain_queue(self.relay_queue)
            _drain_queue(self.json_queue)
            self.json_bytes.clear()

    def _update_rts_state(self):
        """Handle rts being set to some value.

        "self.rts = value" has been executed, for some value. This may not
        have changed the value.
        """
        # We never set rts in our tests, so this doesn't get executed.
        pass  # pragma: no cover

    def _update_dtr_state(self):
        """Handle dtr being set to some value.

        "self.dtr = value" has been executed, for some value. This may not
        have changed the value.
        """
        # We never set dtr in our tests, so this doesn't get executed.
        pass  # pragma: no cover

    def _update_break_state(self):
        """Handle break_condition being set to some value.

        "self.break_condition = value" has been executed, for some value.
        This may not have changed the value.
        Note that break_condition is set and then cleared by send_break().
        """
        # We never set break_condition in our tests, so this doesn't get executed.
        pass  # pragma: no cover

    # --------------------------------------------------------------------------
    # Internal (non-standard) methods.

    def _params_from_url(self, url):
        """Extract various params from the URL."""
        expected = 'expected a string in the form "arduinosimulator://[?board=<name>]"'
        parts = urllib.parse.urlparse(url)

        # Unless we force things (break the normal protocol), scheme will always
        # be 'arduinosimulator'.
        if parts.scheme != 'arduinosimulator':
            raise Exception(expected + ': got scheme {!r}'.format(parts.scheme))  # pragma: no cover
        int_param_names = {'chunk_size', 'read_buffer_size', 'write_buffer_size'}
        params = {}
        for option, values in urllib.parse.parse_qs(parts.query, True).items():
            if option == 'board' and len(values) == 1:
                params[option] = values[0]
            elif option == 'name' and len(values) == 1:
                # This makes it easier for tests to confirm the right serial device has
                # been opened.
                self.name = values[0]
            elif option in int_param_names and len(values) == 1:
                params[option] = int(values[0])
            else:
                raise Exception(expected + ': unknown param {!r}'.format(option))  # pragma: no cover
        return params

    def _create_simulator(self, params):
        board = params.get('board', 'telemetry')
        if board == 'telemetry':
            message = from_json("""
                {
                    "name":"telemetry_board",
                    "ver":"2017-09-23",
                    "power": {
                        "computer":1,
                        "fan":1,
                        "mount":1,
                        "cameras":1,
                        "weather":1,
                        "main":1
                    },
                    "current": {"main":387,"fan":28,"mount":34,"cameras":27},
                    "amps": {"main":1083.60,"fan":50.40,"mount":61.20,"cameras":27.00},
                    "humidity":42.60,
                    "temperature":[13.01,12.81,19.75],
                    "temp_00":15.50
                }
                """)
        elif board == 'camera':
            message = from_json("""
                {
                    "name":"camera_board",
                    "inputs":6,
                    "camera_00":1,
                    "camera_01":1,
                    "accelerometer": {"x":-7.02, "y":6.95, "z":1.70, "o": 6},
                    "humidity":59.60,
                    "temperature":[13.01,12.81,19.75],
                    "temp_00":12.50
                }
                """)
        elif board == 'json_object':
            # Produce an output that is json, but not what we expect
            message = {}
        else:
            raise Exception('Unknown board: {}'.format(board))  # pragma: no cover

        # The elements of these queues are of type bytes. This means we aren't fully controlling
        # the baudrate unless the chunk_size is 1, but that should be OK.
        chunk_size = params.get('chunk_size', 20)
        self.json_queue = queue.Queue(maxsize=params.get('read_buffer_size', 10000))
        self.relay_queue = queue.Queue(maxsize=params.get('write_buffer_size', 100))

        self.device_simulator = ArduinoSimulator(message, self.relay_queue, self.json_queue,
                                                 chunk_size, self.stop)


Serial = FakeArduinoSerialHandler



import pytest
import requests

from astropy import units as u

from panoptes.utils import serializers
from panoptes.utils.config.client import get_config
from panoptes.utils.config.client import set_config


def test_config_client(dynamic_config_server, config_port):
    assert isinstance(get_config(port=config_port), dict)

    assert set_config('location.horizon', 47 * u.degree,
                      port=config_port) == {'location.horizon': 47 * u.degree}

    # With parsing
    assert get_config('location.horizon', port=config_port) == 47 * u.degree

    # Without parsing
    assert get_config('location.horizon', port=config_port, parse=False) == '47.0 deg'


def test_config_client_bad(dynamic_config_server, config_port, caplog):
    # Bad host will return `None` but also throw error
    assert set_config('foo', 42, host='foobaz') is None
    assert caplog.records[-1].levelname == "INFO"
    assert caplog.records[-1].message.startswith("Problem with set_config")

    # Bad host will return `None` but also throw error
    assert get_config('foo', host='foobaz') is None
    assert caplog.records[-1].levelname == "INFO"
    assert caplog.records[-1].message.startswith("Problem with get_config")


def test_config_reset(dynamic_config_server, config_port, config_host):
    # Check we are at default.
    assert get_config('location.horizon', port=config_port) == 30 * u.degree

    # Set to new value.
    set_config_return = set_config('location.horizon', 47 * u.degree, port=config_port)
    assert set_config_return == {'location.horizon': 47 * u.degree}

    # Check we have changed.
    assert get_config('location.horizon', port=config_port) == 47 * u.degree

    # Reset config
    url = f'http://{config_host}:{config_port}/reset-config'
    response = requests.post(url,
                             data=serializers.to_json({'reset': True}),
                             headers={'Content-Type': 'application/json'}
                             )
    assert response.ok

    # Check we are at default again.
    assert get_config('location.horizon', port=config_port) == 30 * u.degree



#!/usr/bin/env python

import os

import click

from panoptes.utils.data.assets import Downloader
from panoptes.utils.logger import logger

DEFAULT_DATA_FOLDER = os.path.expandvars("$PANDIR/astrometry/data")


@click.command()
@click.option('--folder',
              default=DEFAULT_DATA_FOLDER,
              help=f'Destination folder for astrometry indices. Default: {DEFAULT_DATA_FOLDER}')
@click.option('--keep-going/--no-keep-going',
              default=True,
              help='Ignore download failures and keep going to other downloads, default True.')
@click.option('--narrow-field/--no-narrow-field',
              default=False,
              help='Download narrow field indices, default False.')
@click.option('--wide-field/--no-wide-field',
              default=False,
              help='Download wide field indices, default False.')
@click.option('--verbose', is_flag=True, default=False, help='Log output to console.')
def main(folder=DEFAULT_DATA_FOLDER,
         keep_going=True,
         narrow_field=False,
         wide_field=False,
         verbose=False):

    if verbose:
        logger.enable('panoptes')
    else:
        logger.remove(0)

    if not os.path.exists(folder):
        logger.warning(f"Warning, data folder {folder} does not exist, will create.")

    # --no_narrow_field is the default, so the the args list below ignores args.no_narrow_field.
    dl = Downloader(
        data_folder=folder,
        keep_going=keep_going,
        narrow_field=narrow_field,
        wide_field=wide_field,
        verbose=verbose)

    success = dl.download_all_files()
    if success:
        logger.success(f'Downloaded all files')

    return success


if __name__ == '__main__':
    main()
